/*
 * Several students plan to cook something in order to have a nice
 * dinner that tastes different than the food offered in the
 * serveries. The students meet at a certain time at a certain
 * place. Of course, some students may be a bit late but since
 * everybody is very polite the students all wait until the group is
 * complete. Once the group is complete, the work is divided and
 * cooking begins. Unfortunately, some students do not like cooking
 * and they decide to not participate in the cooking process. Instead
 * they immediately sit down at the table waiting for the food to
 * arrive, which happens once the cooking students have finish cooking
 * and sit down at the table as well. (How the food actually gets on
 * the table is not relevant here.) When all students are sitting at
 * the table, all students start to eat. The students discuss that it
 * is not fair that some students did not participate in the cooking
 * process and they finally decide that the non-cooking students are
 * in charge to cleanup the kitchen.  After the last student has
 * finished eating, the non-cooking students start cleaning up the
 * kitchen while the other students simply leave.
 */

#define _POSIX_C_SOURCE 200809L

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <pthread.h>

static const char *progname = "cooking";

#define STUDENT_STATE_IDLE	' '
#define STUDENT_STATE_ARRIVING	'A'
#define STUDENT_STATE_COOKING	'C'
#define STUDENT_STATE_SITTING	'S'
#define STUDENT_STATE_EATING	'E'
#define STUDENT_STATE_WAITING	'W'
#define STUDENT_STATE_CLEANING	'D'

#define STUDENT_TYPE_COOK	'C'
#define STUDENT_TYPE_LAZY	'L'

typedef struct student {
    int id;			/* the identification of a student */
    int state;			/* the student's current state */
    int type;			/* the type of student (cook, lazy) */
    pthread_t thread;		/* the thread representing the student */
} student_t;

static pthread_barrier_t all_students_barrier;
static pthread_barrier_t cooking_students_barrier;
static pthread_barrier_t cleaning_students_barrier;

static void
state(student_t *student, int new_state)
{
    fprintf(stderr, "%3d: %c -> %c\n", student->id, student->state, new_state);
    student->state = new_state;
}

static void*
student(void *arg)
{
    student_t *student = arg;
    
    state(student, STUDENT_STATE_ARRIVING);
    (void) pthread_barrier_wait(&all_students_barrier);
    if (student->type == STUDENT_TYPE_COOK) {
	state(student, STUDENT_STATE_COOKING);
	(void) pthread_barrier_wait(&cooking_students_barrier);
    }
    state(student, STUDENT_STATE_SITTING);
    (void) pthread_barrier_wait(&all_students_barrier);
    state(student, STUDENT_STATE_EATING);
    (void) pthread_barrier_wait(&all_students_barrier);
    if (student->type != STUDENT_TYPE_COOK) {
	state(student, STUDENT_STATE_CLEANING);
	(void) pthread_barrier_wait(&cleaning_students_barrier);
    }
    state(student, STUDENT_STATE_IDLE);
    return NULL;
}

static void
run_threads(int n, int c)
{
    student_t students[n];
    int err;

    assert(n >= 0 && c >=0 && c <= n);
    
    memset(students, 0, sizeof(students));

    err = pthread_barrier_init(&all_students_barrier, NULL, n);
    if (err) {
	fprintf(stderr, "%s: %s(): %s\n", progname, __func__, strerror(err));
	exit(EXIT_FAILURE);
    }
    err = pthread_barrier_init(&cooking_students_barrier, NULL, c);
    if (err) {
	fprintf(stderr, "%s: %s(): %s\n", progname, __func__, strerror(err));
	exit(EXIT_FAILURE);
    }
    err = pthread_barrier_init(&cleaning_students_barrier, NULL, n-c);
    if (err) {
	fprintf(stderr, "%s: %s(): %s\n", progname, __func__, strerror(err));
	exit(EXIT_FAILURE);
    }

    for (int i = 0; i < n; i++) {
        err = pthread_create(&(students[i].thread), NULL,
			     student, &students[i]);
        if (err != 0) {
            fprintf(stderr, "%s: %s(): unable to create thread %d: %s\n",
                    progname, __func__, i, strerror(err));
	    exit(EXIT_FAILURE);
        }
	students[i].id = i;
	students[i].state = STUDENT_STATE_IDLE;
	students[i].type = (i < c) ? STUDENT_TYPE_COOK : STUDENT_TYPE_LAZY;
    }

    for (int i = 0; i < n; i++) {
        if (students[i].thread) {
	    err = pthread_join(students[i].thread, NULL);
	    if (err != 0) {
		fprintf(stderr, "%s: %s(): unable to join thread %d: %s\n",
			progname, __func__, i, strerror(err));
	    }
	}
    }

    (void) pthread_barrier_destroy(&all_students_barrier);
    (void) pthread_barrier_destroy(&cooking_students_barrier);
    (void) pthread_barrier_destroy(&cleaning_students_barrier);
}

int main(int argc, char *argv[])
{
    const char *usage = "Usage: %s [-n students] [-c cooks] [-h]\n";
    int n = 5, c = 1;
    int o, x;

    while ((o = getopt(argc, argv, "n:c:h")) >= 0) {
	switch (o) {
	case 'n':
	    x = atoi(optarg);
	    if (x <= 0) {
		fprintf(stderr, "%s: illegal number: %s\n", progname, optarg);
		exit(EXIT_FAILURE);
	    }
	    n = x;
	    break;
	case 'c':
	    x = atoi(optarg);
	    if (x <= 0) {
		fprintf(stderr, "%s: illegal number: %s\n", progname, optarg);
		exit(EXIT_FAILURE);
	    }
	    c = x;
	    break;
	case 'h':
	    printf(usage, progname);
	    exit(EXIT_SUCCESS);
	}
    }

    if (c > n) {
	fprintf(stderr, "%s: %d students can't have %d cooks\n",
		progname, n, c);
	exit(EXIT_FAILURE);
    }

    run_threads(n, c);
    
    return EXIT_SUCCESS;
}
